package lsedit;

import java.util.Enumeration;
import java.util.Vector;
import java.util.Date;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.Toolkit;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import java.io.File;
import java.io.FileOutputStream;
import java.io.BufferedOutputStream;
import java.io.PrintStream;

import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JPanel;

class ClusterInterface extends JDialog implements ActionListener {


    static String			m_default_command       = "aa";
	static String			m_default_export        = "/temp/ta";
	static String			m_default_import        = "/temp/ca";
	static boolean			m_default_delete_export = false;
	static boolean			m_default_delete_import = false;
	static boolean			m_default_leaves        = true;
	static boolean			m_default_feedback      = true;

	protected LandscapeEditorCore	m_ls;

	protected JTextField	m_command;
	protected JTextField	m_export;
	protected JTextField	m_import;		

	protected JCheckBox		m_delete_export;
	protected JCheckBox		m_delete_import;
	protected JCheckBox		m_leaves;
	protected JCheckBox		m_feedback;

	protected JLabel		m_message;

	protected JButton		m_ok;
	protected JButton		m_default;
	protected JButton		m_cancel;

	// Working registers
	protected Diagram		m_dg;
	
	protected void log(String message)
	{
		if (m_feedback.isSelected()) {
			Date date = new Date();

			System.out.println(date.toLocaleString() + ": " + message);
	}	}

	protected void message(String string)
	{
		m_message.setText(string);
		log(string);
	}

	protected EntityInstance parentOfSet(Vector	set)
	{
		EntityInstance	e, parent, parent1;
		Enumeration		en;

		parent = null;
		for (en = set.elements(); en.hasMoreElements(); ) {
			e     = (EntityInstance) en.nextElement();
			parent1 = e.getContainedBy();
			if (parent == null) {
				parent = parent1;
			} else if (parent != parent1) {
				parent = null;
				break;
		}	}
		return(parent);
	}

	protected void cluster(EntityInstance parent, EntityInstance e)
	{
		Enumeration		en;
		EntityInstance	e1    = null;
		EntityInstance	child = null;

		for (en = e.getChildren(); en.hasMoreElements(); ) {
			if (child == null) {
				e1 = m_dg.updateImportEntity(parent, e, null);
			}
			child = (EntityInstance) en.nextElement();
			cluster(e1, child);
		}
		if (child == null) {
			// This is a leaf node in the import
			String			id    = e.getId();
			EntityInstance	match = m_dg.getCache(id);
			e1 = m_dg.updateImportEntity(parent, e, match);
			e1.nandMark(EntityInstance.SPRING_MARK);
		}
		// Signal that parent being used
		parent.nandMark(EntityInstance.SPRING_MARK);
	}

	protected boolean doit()
	{
		Diagram			dg = m_ls.getDiagram();
		EntityInstance	parent;
		String			export = m_export.getText();



		if (export.length() == 0) {
			message("Please specify an export file to write to");
			return(false);
		}

		// get user's selection of boxes to be laid out

		Vector masterBoxes = dg.getGroup();
		if (masterBoxes == null) {
		  Util.beep();
		  message("No group selected");
		  return(false);
		}

		parent = parentOfSet(masterBoxes);
		if (parent == null) {
			message("Selected items do not have the same parent entity");
			return(false);
		}

		// Emit the export 

		File				exportfile;
		Vector				selectedBoxes;
		Enumeration			en, en1;
		EntityInstance		e, e1;
		RelationInstance	ri;
		RelationClass		rc;
		int					cIndex;
		boolean				leaves = m_leaves.isSelected();

		if (!leaves) {
			selectedBoxes = masterBoxes;
		} else {
			selectedBoxes = new Vector();
			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.gatherLeaves(selectedBoxes);
		}	}

		log("Exporting " + export);

		try {
			exportfile                 = new File(export);
			FileOutputStream	 os    = new FileOutputStream(exportfile);
			BufferedOutputStream bos   = new BufferedOutputStream(os);
			PrintStream			 ps    = new PrintStream(bos);

			ps.print("// Clusting TA file written by LSEdit " + Version.Number() + "\n\n");

			dg.writeSchemeTuples(ps);
			dg.writeSchemeAttributes(ps);
			ps.print("\n\nFACT TUPLE :\n\n");
			
			for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.writeInstance(ps);
				e.orMark(EntityInstance.SPRING_MARK);
			}

			for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				if (leaves) {
					// Use real edges to determine relationships
					for (en1 = e.srcRelationElements(); en1.hasMoreElements(); ) {
						ri = (RelationInstance) en1.nextElement();
						rc = ri.getRelationClass();
						// Consider only visible edges when drawing layout
						if (!rc.isClassVisible()) {
							continue;
						}
						e1 = ri.getDst();
						if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
							continue;
						}
						ri.writeRelation(ps); 

					}
				} else {
					// Use lifted edges to determine relationships
					for (en = e.srcLiftedRelationElements(); en.hasMoreElements(); ) {
						ri = (RelationInstance) en.nextElement();
						rc = ri.getRelationClass();
						// Consider only visible edges when drawing layout
						if (!rc.isClassVisible()) {
							continue;
						}
						e1 = ri.getDrawDst();
						if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
							continue;
						}
						ri.writeRelation(ps); 
			}	}	}

			ps.print("\n\nFACT ATTRIBUTE :\n\n"); 

			cIndex = dg.getCIndex();

			for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.writeAttribute(ps, cIndex);
				if (leaves) {
					// Use real edges to determine relationships
					for (en1 = e.srcRelationElements(); en1.hasMoreElements(); ) {
						ri = (RelationInstance) en1.nextElement();
						rc = ri.getRelationClass();
						// Consider only visible edges when drawing layout
						if (!rc.isClassVisible()) {
							continue;
						}
						e1 = ri.getDst();
						if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
							continue;
						}
						ri.writeAttributes(ps); 

					}
				} else {
					// Use lifted edges to determine relationships
					for (en1 = e.srcLiftedRelationElements(); en1.hasMoreElements(); ) {
						ri = (RelationInstance) en1.nextElement();
						rc = ri.getRelationClass();
						// Consider only visible edges when drawing layout
						if (!rc.isClassVisible()) {
							continue;
						}
						e1 = ri.getDrawDst();
						if (!e1.isMarked(EntityInstance.SPRING_MARK)) {
							continue;
						}
						ri.writeAttributes(ps); 
			}	}	}

			ps.flush();
			ps.close();

			log("Export written");
		}
		catch(Exception error) {
			message("Exception writing '" + export + "' " + error);
			return(false);
		}

		String command = m_command.getText();

		if (command.length() == 0) {
			message("Export written to '" + export + "'. No command specified");
			return(false);
		}

		// Execute the command
		log("Executing [" + command + "]");

		if (m_delete_export.isSelected()) {
			try {
				if (!exportfile.delete()) {
					message("Unable to delete '" + exportfile + "'");
				} else {
					log("Deleted " + exportfile);
			}	}
			catch (Exception error) {
				message("Exception deleting '" + exportfile + "' " + error);
		}	}

		log("Command executed");

		// Import the result

		String	importname = m_import.getText();
		File	importFile;

		log("Importing '" + importname);

		Ta		ta     = new Ta(m_ls);
		String	result = ta.loadTA(importname, null);

		if (result != null) {
			message(result);
			return(false);
		}

		log("Import loaded");

		if (m_delete_import.isSelected()) {
			try {
				File importfile = new File(importname);

				if (!importfile.delete()) {
					message("Unable to delete '" + importfile + "'");
				} else {
					message("Deleted " + importfile);
				}
			} 
			catch (Exception error) {
				message("Exception deleting '" + importname + "' " + error);
		}	}

		m_dg     = dg;

		log("Reclustering");

		if (leaves) {
			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				e.orMark(EntityInstance.SPRING_MARK);
		}	}

		dg.beginUndoRedo("Cluster Interface");

		EntityInstance root = ta.getRootInstance();
		EntityInstance child;

		for (en = root.getChildren(); en.hasMoreElements(); ) {
			child = (EntityInstance) en.nextElement();
			cluster(parent /* My TA */, child /* Other TA */);
		}

		// Make sure that we leave things as entered
		for (en = selectedBoxes.elements(); en.hasMoreElements(); ) {
			e = (EntityInstance) en.nextElement();
			e.nandMark(EntityInstance.SPRING_MARK);
		}

		if (leaves) {
			log("Removing redundant containers");

			/* These items were marked and are not things in selectedBoxes
			 * and have had nothing added back into/under them if SPRING_MARK
			 * still set
			 */

			for (en = masterBoxes.elements(); en.hasMoreElements(); ) {
				e = (EntityInstance) en.nextElement();
				if (e.isMarked(EntityInstance.SPRING_MARK)) {
					dg.updateCutEntity(e);
		}	}	}

		dg.endUndoRedo();

		dg.redrawDiagram();

		message("Clustering operation complete");

		return(true);
	}

	public ClusterInterface(LandscapeEditorCore ls)
	{
		super(ls.getFrame(), "Invoke external clustering tool", true);

		Container			contentPane;
		Font				font, bold;
		JLabel				label;
		JTextField			textfield;
		JCheckBox			checkbox;

		m_ls = ls;
		font = FontCache.getDialogFont();
		bold = font.deriveFont(Font.BOLD);

		setForeground(ColorCache.get(0,0,0));
		setBackground(ColorCache.get(192,192,192));
		setFont(font);

		JPanel topPanel    = new JPanel();
		JPanel labelPanel  = new JPanel();
		JPanel valuePanel  = new JPanel();
	

		GridLayout gridLayout;

		topPanel.setLayout(new BorderLayout());
		gridLayout = new GridLayout(3, 1, 0, 10);
		labelPanel.setLayout(gridLayout);
		gridLayout = new GridLayout(3, 1, 0, 10);
		valuePanel.setLayout(gridLayout);
	
		label = new JLabel("Command to be executed:", JLabel.RIGHT);
		label.setFont(bold);
		labelPanel.add(label);
		m_command = textfield = new JTextField(m_default_command, 80);
		textfield.setFont(font);
		valuePanel.add(textfield);

		label = new JLabel("Exported file used by command:", JLabel.RIGHT);
		label.setFont(bold);
		labelPanel.add(label);
		m_export = textfield = new JTextField(m_default_export, 80);
		textfield.setFont(font);
		valuePanel.add(textfield);

		label = new JLabel("Imported file created by command:", JLabel.RIGHT);
		label.setFont(bold);
		labelPanel.add(label);
		m_import = textfield = new JTextField(m_default_import, 80);
		textfield.setFont(font);
		valuePanel.add(textfield);

		JPanel optionsPanel = new JPanel();
		optionsPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

		m_delete_export = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Delete export file");
		checkbox.setSelected(m_default_delete_export);
		optionsPanel.add(checkbox);
	
		m_delete_import = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Delete import file");
		checkbox.setSelected(m_default_delete_import);
		optionsPanel.add(checkbox);

		m_leaves = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Cluster leaves");
		checkbox.setSelected(m_default_leaves);
		optionsPanel.add(checkbox);
	
		m_feedback = checkbox = new JCheckBox();
		checkbox.setFont(font);
		checkbox.setText("Feedback");
		checkbox.setSelected(m_default_feedback);
		optionsPanel.add(checkbox);

		topPanel.add(BorderLayout.WEST, labelPanel);
		topPanel.add(BorderLayout.EAST, valuePanel);
		topPanel.add(BorderLayout.SOUTH, optionsPanel);

		contentPane = getContentPane();
		contentPane.add( BorderLayout.NORTH, topPanel );

		m_message = new JLabel(" ", JLabel.CENTER);
		m_message.setFont(font);
		m_message.setForeground(Color.RED);
		m_message.setSize(400,50);
		m_message.setPreferredSize(new Dimension(400,50));
		contentPane.add( BorderLayout.CENTER, m_message);

		// --------------
		// Use a FlowLayout to center the button and give it margins.

		JPanel bottomPanel = new JPanel();

		bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

		m_ok = new JButton("Ok");
		m_ok.setFont(bold);
		bottomPanel.add(m_ok);
		m_ok.addActionListener(this);

		m_default = new JButton("Default");
		m_default.setFont(bold);
		bottomPanel.add(m_default);
		m_default.addActionListener(this);

		m_cancel = new JButton("Cancel");
		m_cancel.setFont(bold);
		bottomPanel.add(m_cancel);
		m_cancel.addActionListener(this);

		contentPane.add( BorderLayout.SOUTH, bottomPanel);

		// Resize the window to the preferred size of its components
		pack();
		show();
	}

	// ActionListener interface

	public void actionPerformed(ActionEvent ev)
	{
		Object	source;


		source = ev.getSource();

		if (source == m_default) {
			m_command.setText(m_default_command);
			m_export.setText(m_default_export);
			m_import.setText(m_default_import);
			m_delete_export.setSelected(m_default_delete_export);
			m_delete_import.setSelected(m_default_delete_import);
			m_feedback.setSelected(m_default_feedback);
			return;
		}

		if (source == m_ok) {
			boolean ret;

			m_ls.setCursor(Cursor.WAIT_CURSOR);
			ret = doit();
			m_ls.setCursor(Cursor.DEFAULT_CURSOR);

			if (!ret) {
				return;
			}
			m_default_command       = m_command.getText();
			m_default_export        = m_export.getText();
			m_default_import        = m_import.getText();
			m_default_delete_export = m_delete_export.isSelected();
			m_default_delete_import = m_delete_import.isSelected();
			m_default_feedback      = m_feedback.isSelected();
		} else if (source != m_cancel) {
			return;
		}
		setVisible(false);
		return;
	}

	public static void execute(LandscapeEditorCore ls)
	{
		ClusterInterface dialog = new ClusterInterface(ls);
		dialog.dispose();
	}
} 





