package lsedit;

import java.util.Enumeration;
import java.util.Vector;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Color;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.FlowLayout;
import java.awt.Toolkit;
import java.awt.FontMetrics;
import java.awt.Font;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JDialog;
import javax.swing.JTextField;
import javax.swing.JOptionPane;
import javax.swing.JLabel;
import javax.swing.JButton;
import javax.swing.JPanel;

public class SpringLayout extends LandscapeLayouter  implements ToolBarEventHandler {


	static protected final int ATTRACTIVE_FORCE = 0;
	static protected final int SPARSE_FACTOR    = 1;
	static protected final int CLIENT_FORCE     = 2;
	static protected final int CLIENT_FACTOR    = 3;
	static protected final int SUPPLIER_FORCE   = 4;
	static protected final int SUPPLIER_FACTOR  = 5;
	static protected final int REPULSIVE_FORCE  = 6;
	static protected final int REPULSIVE_DIAMETER = 7;

	// Integer values from here on

	static protected final int ITERATIONS       = 8;
	static protected final int TIMEOUT          = 9;

	static protected final double default_attractive_force = 0.05;
	static protected final double default_sparse_factor    = 1.0;
	static protected final double default_client_force     = 0.05;
	static protected final double default_client_factor    = 0.5;
	static protected final double default_supplier_force   = 0.05;
	static protected final double default_supplier_factor  = 0.5;
	static protected final double default_repulsive_force  = 0.01;
	static protected final double default_repulsive_diameter = 0.75;
	static protected final int    default_iterations       = 1000;
	static protected final int    default_timeout          = 300;	/* 5 minutes */

	protected double	m_attractive_force = default_attractive_force;
	protected double	m_sparse_factor    = default_sparse_factor;
	protected double	m_client_force     = default_client_force;
	protected double	m_client_factor    = default_client_factor;
	protected double	m_supplier_force   = default_supplier_force;
	protected double	m_supplier_factor  = default_supplier_factor;
	protected double	m_repulsive_force  = default_repulsive_force;
	protected double	m_repulsive_diameter = default_repulsive_diameter;
	protected int		m_iterations       = default_iterations;
	protected int		m_timeout          = default_timeout;

	protected static String[]	parameter_names = {
	 "Related force",
	 "Sparseness factor",
	 "Client force",
	 "Client factor",
	 "Supplier force",
	 "Supplier factor",
	 "Unrelated force",
	 "Repulsive range",
	 "Iterations",
	 "Time out"
	};

	class SpringConfigure extends JDialog implements ActionListener {


		protected JTextField[]		m_textfields;

		protected JLabel			m_message;

		protected JButton			m_ok;
		protected JButton			m_default;
		protected JButton			m_cancel;
		protected JButton			m_help;
		
		public SpringConfigure(SpringLayout layout)
		{
			super(layout.getLs().getFrame(), layout.getName() + " Configuration", true);

			Container			contentPane;
			Font				font, bold;
			JLabel				label;
			int					i, size;
			String				string;
			JTextField			textfield;

			size   = parameter_names.length;
			font   = FontCache.getDialogFont();
			bold   = font.deriveFont(Font.BOLD);

			setForeground(ColorCache.get(0,0,0));
			setBackground(ColorCache.get(192,192,192));
			setFont(font);

			JPanel topPanel    = new JPanel();
			JPanel labelPanel1 = new JPanel();
			JPanel valuePanel1 = new JPanel();
			JPanel labelPanel2 = new JPanel();
			JPanel valuePanel2 = new JPanel();
			JPanel labelPanel, valuePanel;

			GridLayout gridLayout;

			topPanel.setLayout(new BorderLayout());
			gridLayout = new GridLayout(5, 1, 0, 10);
			labelPanel1.setLayout(gridLayout);
			gridLayout = new GridLayout(5, 1, 0, 10);
			valuePanel1.setLayout(gridLayout);
			gridLayout = new GridLayout(5, 1, 0, 10);
			labelPanel2.setLayout(gridLayout);
			gridLayout = new GridLayout(5, 1, 0, 10);
			valuePanel2.setLayout(gridLayout);

			m_textfields = new JTextField[size];

			for (i = 0; i < size; ++i) {
				if ((i % 2) == 0) {
					labelPanel = labelPanel1;
					valuePanel = valuePanel1;
				} else {
					labelPanel = labelPanel2;
					valuePanel = valuePanel2;
				}

				label = new JLabel(parameter_names[i], JLabel.RIGHT);
				label.setFont(bold);
				labelPanel.add(label);

				string = "";
				switch (i) {
				case ATTRACTIVE_FORCE:
					string += m_attractive_force;
					break;
				case SPARSE_FACTOR:
					string += m_sparse_factor;
					break;
				case CLIENT_FORCE:
					string += m_client_force;
					break;
				case CLIENT_FACTOR:
					string += m_client_factor;
					break;
				case SUPPLIER_FORCE:
					string += m_supplier_force;
					break;
				case SUPPLIER_FACTOR:
					string += m_supplier_factor;
					break;
				case REPULSIVE_FORCE:
					string += m_repulsive_force;
					break;
				case REPULSIVE_DIAMETER:
					string += m_repulsive_diameter;
					break;
				case ITERATIONS:
					string += m_iterations;
					break;
				case TIMEOUT:
					string += m_timeout;
				}

				m_textfields[i] = textfield = new JTextField(string, 4);
				textfield.setFont(font);
				textfield.addActionListener(this);
				valuePanel.add(textfield);
			}

			JPanel leftPanel   = new JPanel();
			leftPanel.setLayout(new BorderLayout());
			leftPanel.add(BorderLayout.WEST, labelPanel1);
			leftPanel.add(BorderLayout.EAST, valuePanel1);

			JPanel rightPanel = new JPanel();
			rightPanel.setLayout(new BorderLayout());
			rightPanel.add(BorderLayout.WEST, labelPanel2);
			rightPanel.add(BorderLayout.EAST, valuePanel2);
	
			topPanel.add(BorderLayout.WEST, leftPanel);
			topPanel.add(BorderLayout.EAST, rightPanel);

			contentPane = getContentPane();
			contentPane.add( BorderLayout.NORTH, topPanel );

			m_message = new JLabel(" ", JLabel.CENTER);
			m_message.setFont(font);
			m_message.setForeground(Color.RED);
			m_message.setSize(400,50);
			m_message.setPreferredSize(new Dimension(400,50));
			contentPane.add( BorderLayout.CENTER, m_message);

			// --------------
			// Use a FlowLayout to center the button and give it margins.

			JPanel bottomPanel = new JPanel();

			bottomPanel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));

			m_ok = new JButton("Ok");
			m_ok.setFont(bold);
			bottomPanel.add(m_ok);
			m_ok.addActionListener(this);

			m_default = new JButton("Default");
			m_default.setFont(bold);
			bottomPanel.add(m_default);
			m_default.addActionListener(this);

			m_cancel = new JButton("Cancel");
			m_cancel.setFont(bold);
			bottomPanel.add(m_cancel);
			m_cancel.addActionListener(this);


			m_help = new JButton("Help");
			m_help.setFont(bold);
			bottomPanel.add(m_help);
			m_help.addActionListener(this);

			contentPane.add( BorderLayout.SOUTH, bottomPanel);

			// Resize the window to the preferred size of its components
			pack();
			show();
		}

		// ActionListener interface

		public void actionPerformed(ActionEvent ev)
		{
			Object		source;
			String		string;
			double[]	values     = {0, 0, 0, 0, 0, 0, 0, 0};
			int[]		ivalues    = {0, 0};
			int			ivalue; 

			source = ev.getSource();

			if (source == m_help) {
				JOptionPane.showMessageDialog(m_ls.getFrame(), 	
				  "Attractive force\n" +
				  "  Scales size of jumps towards ideal distance for related nodes.\n" +
				  "  Actual jump is proportional to difference from ideal distance\n" +
				  "Repulsive force\n" +
				  "  Scales size of jumps towards repulsive diameter for unrelated nodes\n" +
				  "  Actual jump is proportional to difference from repulsive diameter\n" +
				  "Client force\n" +
				  "  Scales size of jumps towards ideal distance from clients\n" +
				  "Supplier force\n" +
				  "  Scales size of jumps towards ideal distance from suppliers\n\n" +
				  "Sparse factor\n" +
				  "  The maximum width/height of any two distinct nodes is multiplied\n" +
				  "  by the sparse factor to give the ideal distance\n" +
				  "Client factor\n" +
				  "  The maximum width/height of any two distinct nodes is multiplied\n" +
				  "  by the client factor to give the ideal distance from clients\n" +
				  "Supplier factor\n" +
				  "  The maximum width/height of any two distinct nodes is multiplied\n" +
				  "  by the supplier factor to give the ideal distance from suppliers\n" +
				  "Repulsive diameter\n" +
				  "  Ideal ratio of distance between unrelated nodes to screen size\n" +
				  "Iterations\n" +
				  "  Number of times to iterate over the algorithm\n" +
				  "Timeout\n" +
				  "  Maximum number of seconds to spend iterating"  
				 	  , "Help", JOptionPane.OK_OPTION);
				return;
			}

			int		size   = parameter_names.length;
			int		i;

			if (source == m_default) {
				for (i = 0; i < size; ++i) {
					string = "";
					switch (i) {
					case ATTRACTIVE_FORCE:
						string += default_attractive_force;
						break;
					case SPARSE_FACTOR:
						string += default_sparse_factor;
						break;
					case CLIENT_FORCE:
						string += default_client_force;
						break;
					case CLIENT_FACTOR:
						string += default_client_factor;
						break;
					case SUPPLIER_FORCE:
						string += default_supplier_force;
						break;
					case SUPPLIER_FACTOR:
						string += default_supplier_factor;
						break;
					case REPULSIVE_FORCE:
						string += default_repulsive_force;
						break;
					case REPULSIVE_DIAMETER:
						string += default_repulsive_diameter;
						break;
					case ITERATIONS:
						string += default_iterations;
						break;
					case TIMEOUT:
						string += default_timeout;
					}
					m_textfields[i].setText(string);
				}
				return;
			}

			for (i = 0; i < size; ++i) {
				if (source == m_textfields[i] || source == m_ok) {
					string = m_textfields[i].getText();
					string.trim();
					switch (i) {
					case ITERATIONS:
					case TIMEOUT:
						try {
							ivalue = Integer.parseInt(string);
							if (ivalue <= 0) {
								m_message.setText(parameter_names[i] + " must be +ve");
								return;
							}
							ivalues[i-ITERATIONS] = ivalue;
						} catch (Throwable exception) {
							m_message.setText(parameter_names[i] + " not an integer value");
							return;
						}
						break;
					default:
						try {
							values[i] = Double.parseDouble(string);
						} catch (Throwable exception) {
							m_message.setText(parameter_names[i] + " not a double precision value");
							return;
			}	}	}	}

			if (source != m_cancel && source != m_ok) {
				return;
			}

			if (source == m_ok) {
				m_attractive_force   = values[ATTRACTIVE_FORCE];
				m_sparse_factor      = values[SPARSE_FACTOR];
				m_client_force       = values[CLIENT_FORCE];
				m_client_factor      = values[CLIENT_FACTOR];
				m_supplier_force     = values[SUPPLIER_FORCE];
				m_supplier_factor    = values[SUPPLIER_FACTOR];
				m_repulsive_force    = values[REPULSIVE_FORCE];
				m_repulsive_diameter = values[REPULSIVE_DIAMETER];
				m_iterations         = ivalues[0];
				m_timeout            = ivalues[1];
			}
			setVisible(false);
			return;
		}
	}

	class SpringNode {
		public EntityInstance	m_e;
		double					m_x;
		double					m_y;
		double					m_xmax;
		double					m_ymax;
		boolean					m_clients;
		boolean					m_suppliers;
	};

	public SpringLayout(LandscapeEditorCore ls) 
	{
		super(ls);
	}

/****************************************************************************/

	public String getName()
	{
		return "Spring";
	}

	public String getMenuLabel() 
	{
		return "Layout using Springs";
	} 

/***************************************************************************/


	public boolean isConfigurable()
	{
		return true;
	}

	public void configure(LandscapeEditorCore ls)
	{
		SpringConfigure configure = new SpringConfigure(this);
		configure.dispose();
	}

	// The doLayout method executes the Network Simplex algorithm
	// Assumption: All boxes selected are in the same container.

	public void doLayout1(Vector selectedBoxes, EntityInstance container) 
	{
		int					size = selectedBoxes.size();
		int					i, j, iteration;
		EntityInstance		e, e1;
		Enumeration			en;
		RelationInstance	ri;
		RelationClass		rc;
		double				width, height, width1, height1;
		double				x, y, x1, y1, xmax, ymax, x1max, y1max, xdiff, radius, radius1, ydiff, f, length, newlength;
		double				ideal_length, ideal_client, ideal_supplier;
		long				timeout;

		switch (size) {
		case 0:
			return;
		case 1:
			e = (EntityInstance) selectedBoxes.firstElement();
			width  = 0.5;
			height = 0.5;
			x      = (1 - width)/2;
			y      = (1 - height)/2;

			e.updateRelLocal(x, y, width, height);
			return;
		}

		SpringNode[]	springNodes = new SpringNode[size];
		SpringNode		springNode, springNode1;
		boolean[][]		related;
		boolean			isTopClients = m_ls.isTopClients();
		boolean			temp;

		related = new boolean[size][];

		for (i = 0; i < size; ++i) {
			springNodes[i]    = springNode = new SpringNode();
			related[i]        = new boolean[size-i];
			springNode.m_e    = e = (EntityInstance) selectedBoxes.elementAt(i);
			springNode.m_x    = e.xRelLocal();
			springNode.m_y    = e.yRelLocal();
			springNode.m_xmax = 1.0 - e.widthRelLocal();
			springNode.m_ymax = 1.0 - e.heightRelLocal();
			springNode.m_clients = false;
			springNode.m_suppliers = false;
			e.orMark(EntityInstance.SPRING_MARK);
		}

		// Compute related and ideal length (max of two times any two widths or two hights -- converts 1.4 on diagonal)

		ideal_length   = 0;
		ideal_client   = 0;
		ideal_supplier = 0;
		for (i = 0; i < size; ++i) {
			springNode        = springNodes[i];
			e                 = springNode.m_e;
			width             = e.widthRelLocal();
			height            = e.heightRelLocal();
			for (en = e.srcLiftedRelationElements(); en.hasMoreElements(); ) {
				ri = (RelationInstance) en.nextElement();
				rc = ri.getRelationClass();
				// Consider only visible edges when drawing layout
				if (rc.isClassVisible()) {
					e1       = ri.getDrawDst();

					if (e1.isMarked(EntityInstance.CLIENT_SUPPLIER)) {
						springNode.m_suppliers = true;
					} 
					if (e1.isMarked(EntityInstance.SPRING_MARK)) {
						for (j = i+1; j < size; ++j) {
							if (e1 == springNodes[j].m_e) {
								related[i][j-i] = true;
								related[i][0]   = true;
								related[j][0]   = true;
								width1          = e1.widthRelLocal();
								height1         = e1.heightRelLocal();
								xdiff           = width + width1;
								if (xdiff > ideal_length) {
									ideal_length = xdiff;
								}
								ydiff           = height + height1;
								if (ydiff > ideal_length) {
									ideal_length = ydiff;
								}
								break;
			}	}	}	}	}	
			for (en = e.dstLiftedRelationElements(); en.hasMoreElements(); ) {
				ri = (RelationInstance) en.nextElement();
				rc = ri.getRelationClass();
				// Consider only visible edges when drawing layout
				if (rc.isClassVisible()) {
					e1       = ri.getDrawSrc();
					if (e1.isMarked(EntityInstance.CLIENT_SUPPLIER)) {
						springNode.m_clients = true;
					} 
					if (e1.isMarked(EntityInstance.SPRING_MARK)) {
						for (j = i+1; j < size; ++j) {
							if (e1 == springNodes[j].m_e) {
								related[i][j-i] = true;
								related[i][0]   = true;
								related[j][0]   = true;
								width1          = e1.widthRelLocal();
								height1         = e1.heightRelLocal();
								xdiff           = width + width1;
								if (xdiff > ideal_length) {
									ideal_length = xdiff;
								}
								ydiff           = height + height1;
								if (ydiff > ideal_length) {
									ideal_length = ydiff;
								}
								break;
			}	}	}	}	}
			if (springNode.m_clients || springNode.m_suppliers) {
				if (springNode.m_clients & springNode.m_suppliers) {
					springNode.m_clients   = false;
					springNode.m_suppliers = false;
				} else {
					if (!isTopClients) {
						temp                   = springNode.m_clients;
						springNode.m_clients   = springNode.m_suppliers;
						springNode.m_suppliers = temp;
			}	}	}
			e.nandMark(EntityInstance.SPRING_MARK);
		}

		ideal_client   = ideal_length * m_client_factor;
		ideal_supplier = ideal_length * m_supplier_factor;
		ideal_length  *= m_sparse_factor;

		timeout = System.currentTimeMillis() + (m_timeout * 1000);

		for (iteration = m_iterations; iteration > 0; --iteration) {

			for (i = 0; i < size; ++i) {
				springNode = springNodes[i];
				x          = springNode.m_x;
				y          = springNode.m_y;
				xmax       = springNode.m_xmax;
				ymax       = springNode.m_ymax;
				if (springNode.m_suppliers || springNode.m_clients) {
					if (springNode.m_suppliers) {
						y  += (ymax - y - ideal_supplier)*m_supplier_force;
					} else {
						y  += (ideal_client - y) * m_client_force;
					}
					if (y > ymax) {
						y = ymax;
					}
					if (y < 0) {
						y = 0;
				}	}

				for (j = i+1; j < size; ++j) {
					springNode1 = springNodes[j];
					x1          = springNode1.m_x;
					y1          = springNode1.m_y;
					x1max       = springNode1.m_xmax;
					y1max       = springNode1.m_ymax;
					xdiff       = x1 - x;
					ydiff       = y1 - y;
					if (xdiff == 0 && ydiff == 0) {
						xdiff = (i%3)-1;
						if (xdiff == 0) {
							ydiff = (j%2)*2-1;
						} else {
							ydiff = (j%3)-1;
					}	} 
					length = Math.sqrt(xdiff*xdiff + ydiff*ydiff);
					if (related[i][j-i]) {
						f = (length - ideal_length)         * m_attractive_force;
					} else {
						f = (length - m_repulsive_diameter) * m_repulsive_force;

					}
					x  += (f * xdiff/length);
					x1 -= (f * xdiff/length);
					y  += (f * ydiff/length);
					y1 -= (f * ydiff/length);
	
					if (x > xmax) {
						x1 += xmax - x;
						x   = xmax;
					}
					if (x1 > x1max) {
						x  += x1max - x1;
						x1  = x1max;
					}
					if (x < 0) {
						x1 -= x;
						x   = 0;
					}
					if (x1 < 0) {
						x  -= x1;
						x1  = 0;
					}
					if (x > xmax) {
						x = xmax;
					}
					if (x1 > x1max) {
						x1 = x1max;
					}
					if (y > ymax) {
						y1 += ymax-y;
						y   = ymax;
					}
					if (y1 > y1max) {
						y  += y1max - y1;
						y1  = y1max;
					}
					if (y < 0) {
						y1 -= y;
						y   = 0;
					}
					if (y1 < 0) {
						y  -= y1;
						y1  = 0;
					}
					if (y > ymax) {
						y = ymax;
					}
					if (y1 > y1max) {
						y1 = y1max;
					}

					springNode1.m_x = x1;
					springNode1.m_y = y1;
				}
				springNode.m_x  = x;
				springNode.m_y  = y;
			}
			if (System.currentTimeMillis() > timeout) {
				System.out.println(toLocaleString() + ": Timeout after " + (m_iterations - iteration) + " iterations");
				break;
		}	}

		for (i = 0; i < size; ++i) {
			springNode = springNodes[i];
			e          = springNode.m_e;
			e.updateRelLocal(springNode.m_x, springNode.m_y, e.widthRelLocal(), e.heightRelLocal());
		}
		
	} // doLayout1


  // The doLayout method executes the Coffman-Graham Layer Assignment
  // algorithm and the Sugiyama algorithm on the boxes selected.
  // Assumption: All boxes selected are in the same container.

	public String doLayout(Diagram dg) 
	{
		EntityInstance parent;

		// get user's selection of boxes to be laid out

		m_ls.setLayouter(this);
		m_ls.setCursor(Cursor.WAIT_CURSOR);

		Vector selectedBoxes = dg.getGroup();
		if (selectedBoxes == null) {
			  Util.beep();
			  return "No group selected";
		}

		String msg = allInDiagram(selectedBoxes);
		if (msg != null) {
			return msg;
		}

		parent = parentOfSet(selectedBoxes);
		if (parent == null) {
			return	"Spring layout requires that all things laid out share same parent";
		}
		dg.beginUndoRedo("Spring layout");
		doLayout1(selectedBoxes, parent);
		dg.endUndoRedo();

		m_ls.setCursor(Cursor.DEFAULT_CURSOR);

		return "Graph redrawn using Spring Layout";
	} // doLayout

	public void processKeyEvent(int key, int modifiers, Object object) 
	{
		Diagram	dg;
		String	rmsg;

		dg = m_ls.getDiagram();
		if (dg != null) {
			rmsg = doLayout(dg);
			m_ls.doFeedback(rmsg);
			dg.rescaleDiagram();
			m_ls.repaintDg();
	}	}

} 





